import trio


async def _azip(*aiterables, fillvalue, stop_any):
    iters = [item.__aiter__() for item in aiterables]
    while True:
        stop_count = 0
        items = [fillvalue] * len(iters)

        async def collect(i, iterator):
            try:
                items[i] = await iterator.__anext__()
            except StopAsyncIteration:
                nonlocal stop_count
                stop_count += 1  # pylint: disable=undefined-variable

        async with trio.open_nursery() as nursery:
            for i, iterator in enumerate(iters):
                nursery.start_soon(collect, i, iterator)

        if stop_count >= (1 if stop_any else len(iters)):
            break
        yield tuple(items)


def azip(*aiterables):
    """async version of zip() with parallel iteration"""
    return _azip(*aiterables, fillvalue=None, stop_any=True)


def azip_longest(*aiterables, fillvalue=None):
    """async version of zip_longest() with parallel iteration"""
    return _azip(*aiterables, fillvalue=fillvalue, stop_any=False)



# pylint: disable=line-too-long,multiple-statements
from contextlib import contextmanager
from typing import TypeVar, Generic, AsyncIterator, Tuple, ContextManager, Callable, overload

import trio

try:
    from trio import lowlevel
except ImportError:  # pragma: no cover
    from trio import hazmat as lowlevel  # type: ignore

from ._ref_counted_default_dict import _RefCountedDefaultDict


class _WaitQueue:
    """Fast replacement for trio.ParkingLot.

    The typical unpark_all() is 2x faster.
    """

    __slots__ = ['tasks']

    def __init__(self):
        self.tasks = set()

    async def park(self):
        task = lowlevel.current_task()
        self.tasks.add(task)

        def abort_fn(_):
            self.tasks.remove(task)
            return lowlevel.Abort.SUCCEEDED

        await lowlevel.wait_task_rescheduled(abort_fn)

    def unpark_all(self):
        for task in self.tasks:
            lowlevel.reschedule(task)
        self.tasks.clear()


def _ANY_TRANSITION(value, old_value):
    return True


def _ANY_VALUE(value):
    return True


_NONE = object()
T = TypeVar('T')
T_OUT = TypeVar('T_OUT')
P = Callable[[T], bool]
P2 = Callable[[T, T], bool]


class _Result:
    __slots__ = ['event', 'value']

    def __init__(self):
        self.event = _WaitQueue()
        self.value = None


# TODO: hash functools.partial by func + args + keywords when possible
class _ValueWrapper:
    """
    If value_or_predicate is not callable, wrap it in a suitable callable, and
    hash by the original value if possible.

    This provides a consistent interface for awaitables, and allows listeners
    of the same plain value to share a wait queue.
    """

    __slots__ = ['value']

    def __new__(cls, value_or_predicate):
        return value_or_predicate if callable(value_or_predicate) else super().__new__(cls)

    def __init__(self, value):
        self.value = value

    def __hash__(self):
        try:
            return hash(self.value)
        except TypeError:
            return super().__hash__()

    def __eq__(self, other):
        return self.value.__eq__(other.value)

    def __call__(self, x, *args):
        return x == self.value


class AsyncValue(Generic[T]):
    """
    Value wrapper offering the ability to wait for a value or transition.

    Synopsis::

        >>> a = AsyncValue(0)  # NOTE: can wrap any type (enum, tuple, ...)
        >>> ...
        >>> a.value = 5   # access underlying value
        >>> ...
        >>> # wait for value match by equality
        >>> await a.wait_value(7)
        >>> ...
        >>> # wait for value match by predicate
        >>> await a.wait_value(lambda v: v > 10)
        >>> ...
        >>> # values via iteration (with eventual consistency)
        >>> async for value a.eventual_values(lambda v: v > 10):
        >>>     ...
        >>> # wait for any transition
        >>> await a.wait_transition()
        >>> ...
        >>> # wait for transition by equality
        >>> await a.wait_transition(14)
        >>> ...
        >>> # wait for transition by predicate
        >>> await a.wait_transition(lambda v, old: v > 10 and old < 0)
        >>> ...
        >>> # repeated transitions via iteration (misses values while blocked)
        >>> async for value, _ in a.transitions(lambda v, old: v > 10 and old < 0):
        >>>     ...

    When using any of the wait methods or iterators in this API, note that
    the value may have changed again before the caller receives control.
    For clarity, the specific value that triggered the wakeup is always
    returned.

    Comparison of eventual_values() and transitions() iterators::

        eventual_values()                      transitions()
        =================                      =============
        • high level & safe                    • low level & requires care
        • evaluated on each value change       • evaluated on each value change
            and when starting loop                 if caller not blocked in the body
        • eventually iterates latest value     • latest value missed if caller is blocked
        • can miss rapid value changes         • can miss rapid value changes
        • condition uses new value only        • condition uses new and/or old value
        • mainly used for sync'ing state       • mainly used for triggering work
        • not subject to user races            • races possible on init and each iteration
                                                   (especially if value changes infrequently)

    Performance note:  assignment to the `value` property typically has O(N)
    complexity, where N is the number of actively waiting tasks.  Shared
    predicates are grouped when possible, reducing N to the number of active
    predicates.
    """

    def __init__(self, value: T):
        self._value = value
        self._level_results = _RefCountedDefaultDict(_Result)
        self._edge_results = _RefCountedDefaultDict(_Result)
        # predicate: AsyncValue
        self._transforms = _RefCountedDefaultDict(lambda: AsyncValue(_NONE))

    def __repr__(self):
        return f"{self.__class__.__name__}({self.value})"

    @property
    def value(self) -> T:
        """The wrapped value"""
        return self._value

    @value.setter
    def value(self, x: T):
        if self._value != x:
            old = self._value
            new = self._value = x
            for f, result in self._level_results.items():
                if f(new):
                    result.value = new
                    result.event.unpark_all()
            for f, result in self._edge_results.items():
                if f(new, old):
                    result.value = (new, old)
                    result.event.unpark_all()
            for f, output in self._transforms.items():
                output.value = f(new)

    @staticmethod
    async def _wait_predicate(result_map, predicate):
        with result_map.open_ref(predicate) as result:
            await result.event.park()
        return result.value

    @overload
    async def wait_value(self, value: T, *, held_for=0.) -> T: ...
    @overload
    async def wait_value(self, predicate: P, *, held_for=0.) -> T: ...
    async def wait_value(self, value_or_predicate, *, held_for=0.):
        """
        Wait until given predicate f(value) is True.

        The predicate is tested immediately and, if false, whenever
        the `value` property changes.

        If a non-callable is provided, it's equivalent to a predicate matching
        the given value.

        If held_for > 0, the predicate must match for that duration from the
        time of the call.  "held" means that the predicate is continuously true.

        returns value which satisfied the predicate
        (when held_for > 0, it's the most recent value)
        """
        predicate = _ValueWrapper(value_or_predicate)
        while True:
            if not predicate(self._value):
                value = await self._wait_predicate(self._level_results, predicate)
            else:
                value = self._value
                await trio.sleep(0)
            if held_for > 0:
                with trio.move_on_after(held_for):
                    await self.wait_value(lambda v: not predicate(v))
                    continue
            break
        return value

    # NOTE: it's hard to make both type check and lint happy with generator overloads
    # https://github.com/python/mypy/issues/10301
    # https://github.com/PyCQA/astroid/issues/1015
    @overload
    async def eventual_values(self, value: T, held_for=0.) -> AsyncIterator[T]: yield self._value
    @overload
    async def eventual_values(self, predicate: P = _ANY_VALUE, held_for=0.) -> AsyncIterator[T]: yield self._value
    async def eventual_values(self, value_or_predicate=_ANY_VALUE, held_for=0.):
        """
        Yield values matching the predicate with eventual consistency

        The initial value will be yielded immediately if it matches the
        predicate.  Subsequent yields occur whenever the `value` property
        changes to a value matching the predicate.  Note that rapid changes as
        well as multiple changes while the caller body is blocked will not all
        be reflected, but eventual consistency is ensured.

        The default predicate will match any value.

        If a non-callable is provided, it's equivalent to a predicate matching
        the given value.

        If held_for > 0, the predicate must match for that duration.
        """
        predicate = _ValueWrapper(value_or_predicate)
        last_value = self._value
        with self._level_results.open_ref(predicate) as result, \
                self._level_results.open_ref(lambda v: v != last_value) as not_last_value, \
                self._level_results.open_ref(lambda v: not predicate(v)) as not_predicate:
            while True:
                if predicate(self._value):
                    last_value = self._value
                else:
                    await result.event.park()
                    last_value = result.value
                if held_for > 0:
                    with trio.move_on_after(held_for):
                        await not_predicate.event.park()
                        continue
                yield last_value
                if self._value == last_value:
                    await not_last_value.event.park()

    # TODO: implement wait_transition() using transitions()
    @overload
    async def wait_transition(self, value: T) -> Tuple[T, T]: ...
    @overload
    async def wait_transition(self, predicate: P2 = _ANY_TRANSITION) -> Tuple[T, T]: ...
    async def wait_transition(self, value_or_predicate=_ANY_TRANSITION):
        """
        Wait until given predicate f(value, old_value) is True.

        The predicate is tested whenever the `value` property changes.
        The default predicate responds to any value change.

        If a non-callable is provided, it's equivalent to a predicate matching
        the given value.

        Note that unlike `wait_value()`, it is easy to have race conditions
        when using `wait_transition()`.  Always consider whether it's possible
        for the initially desired transition to have already occurred due to
        task scheduling order, etc.

        returns (value, old_value) which satisfied the predicate
        """
        return await self._wait_predicate(self._edge_results, _ValueWrapper(value_or_predicate))

    @overload
    async def transitions(self, value: T) -> AsyncIterator[Tuple[T, T]]: yield (self._value, self._value)
    @overload
    async def transitions(self, predicate: P2 = _ANY_TRANSITION) -> AsyncIterator[Tuple[T, T]]: yield (self._value, self._value)
    async def transitions(self, value_or_predicate=_ANY_TRANSITION):
        """
        Yield (value, old_value) for transitions matching the predicate

        Transitions that happen during the body of the loop are discarded.

        The iteration::

            >>> async for value, old_value in async_value.transitions(...)
            >>>     ...

        is equivalent to::

            >>> while True:
            >>>     value, old_value = await async_value.wait_transition(...)
            >>>     ...

        Unlike the `eventual_values()` iterator, use of the `transitions()` is
        prone to races when entering the loop.  Always consider whether it's
        possible for the desired transition to have already occurred due to
        task scheduling order, etc.
        """

        # Note this is not simply `while True: await wait_transition(...)`,
        # in order to maintain a ref on the predicate throughout the loop.
        predicate = _ValueWrapper(value_or_predicate)
        with self._edge_results.open_ref(predicate) as result:
            while True:
                await result.event.park()
                yield result.value

    # TODO: make the output's `value` read-only somehow
    def open_transform(self, function: Callable[[T], T_OUT]) \
            -> ContextManager['AsyncValue[T_OUT]']:
        """Yield a derived AsyncValue with the given transform applied

        Synopsis::

        >>> x = AsyncValue(1)
        >>> with x.open_transform(lambda val: val * 2) as y:
        >>>     assert y.value == 2
        >>>     x.value = 10
        >>>     assert y.value == 20
        """
        # type hint workaround for https://youtrack.jetbrains.com/issue/PY-36444
        return self._open_transform(function)

    @contextmanager
    def _open_transform(self, function):
        with self._transforms.open_ref(function) as output:
            if output.value is _NONE:
                output.value = function(self.value)
            yield output



import logging
from collections import defaultdict

import trio

logger = logging.getLogger(__name__)

RATE_MEASURE_PERIOD = .25


class TaskStats(trio.abc.Instrument):
    """Trio scheduler Instrument which logs various task stats at termination.

    Includes max task wait time, slow task steps, and highest task schedule
    rate.
    """
    # TODO: auto thresholds (e.g. > 98th percentile)
    def __init__(self, *, slow_task_threshold=.01, high_rate_task_threshold=100,
                 current_time=trio.current_time):
        """
        :param slow_task_threshold: tasks with steps higher than this duration
            (seconds) are reported
        :param high_rate_task_threshold: tasks scheduling higher than this rate
            are reported
        :param current_time: function used to retrieve the current time
        """
        super().__init__()
        self.slow_task_threshold = slow_task_threshold
        self.high_rate_task_threshold = high_rate_task_threshold
        self.current_time = current_time
        self.scheduled_start = {}  # task: start_time
        self.max_wait = 0
        self.task_step_start = None
        self.slow_task_steps = defaultdict(list)  # name: dt_list
        self.schedule_counts = defaultdict(int)  # name: count
        self.rate_start = 0
        self.high_schedule_rates = defaultdict(float)  # name: max_rate

    def task_scheduled(self, task):
        t = self.current_time()
        self.scheduled_start[task] = t
        if t - self.rate_start > RATE_MEASURE_PERIOD:
            for name, count in self.schedule_counts.items():
                rate = count / RATE_MEASURE_PERIOD
                if rate >= self.high_rate_task_threshold:
                    self.high_schedule_rates[name] = max(self.high_schedule_rates[name], rate)
            self.rate_start = t
            self.schedule_counts.clear()
        self.schedule_counts[task.name] += 1

    def before_task_step(self, task):
        t = self.current_time()
        start = self.scheduled_start.pop(task, None)
        if start:
            dt = t - start
            self.max_wait = max(self.max_wait, dt)
            self.task_step_start = t

    def after_task_step(self, task):
        start = self.task_step_start
        if start:
            dt = self.current_time() - start
            if dt > self.slow_task_threshold:
                self.slow_task_steps[task.name].append(dt)
            self.task_step_start = None

    def after_run(self):
        logger.info(f'max task wait time: {self.max_wait * 1000:.2f} ms')
        if self.slow_task_steps:
            text = [f'slow task step events (> {self.slow_task_threshold * 1000:.0f} ms):']
            for name, dt_list in sorted(self.slow_task_steps.items(),
                                        key=lambda item: max(item[1]), reverse=True):
                dt_text = ', '.join(f'{dt * 1000:.0f}ms'
                                    for dt in sorted(dt_list, reverse=True))
                text.append(f'  {name}: {dt_text}')
            logger.info('\n'.join(text))
        if self.high_schedule_rates:
            text = ['high task schedule rates:']
            for name, rate in sorted(self.high_schedule_rates.items(),
                                     key=lambda item: item[1], reverse=True):
                text.append(f'  {name}: {rate:.0f} Hz')
            logger.info('\n'.join(text))



import trio

async def periodic(period):
    """Yield `(elapsed_time, delta_time)` with an interval of `period` seconds.

    For example, to loop indefinitely with a period of 1 second, accounting
    for the time taken in the loop itself::

        async for _ in periodic(1):
            ...

    In the case of overrun, the next iteration begins immediately.

    On the first iteration, `delta_time` will be `None`.
    """
    t0 = trio.current_time()
    t_last, t_start = None, t0
    while True:
        yield t_start - t0, t_start - t_last if t_last is not None else None
        await trio.sleep_until(t_start + period)
        t_last, t_start = t_start, trio.current_time()



from ._version import __version__
from ._async_bool import AsyncBool
from ._async_itertools import azip, azip_longest
from ._async_value import AsyncValue
from ._awaitables import wait_all, wait_any
from ._cancel_scopes import move_on_when, run_and_cancelling
from ._compose_values import compose_values
from ._exceptions import defer_to_cancelled, multi_error_defer_to
from ._iterators import iter_fail_after, iter_move_on_after
from ._periodic import periodic
from ._repeated_event import RepeatedEvent
from ._task_stats import TaskStats
from ._trio_async_generator import trio_async_generator

def _metadata_fix():
    # don't do this for Sphinx case because it breaks "bysource" member ordering
    import sys
    if 'sphinx' in sys.modules:  # pragma: no cover
        return

    for name, value in globals().items():
        if not name.startswith('_'):
            value.__module__ = __name__

_metadata_fix()



from collections.abc import AsyncIterator

import trio


class iter_move_on_after(AsyncIterator):
    """async iterator adapter that stops if an iteration exceeds timeout

    The timeout is a duration in seconds.

    Synopsis::

        async for v in iter_move_on_after(5, async_value.eventual_values()):
            ...
    """

    __slots__ = ['_ait', '_timeout']

    def __init__(self, timeout, ait):
        self._ait = ait
        self._timeout = timeout

    async def __anext__(self):
        with trio.move_on_after(self._timeout):
            x = await self._ait.__anext__()
            return x
        raise StopAsyncIteration


class iter_fail_after(AsyncIterator):
    """async iterator adapter that raises trio.TooSlowError if an iteration exceeds timeout

    The timeout is a duration in seconds.

    Synopsis::

        async for v in iter_fail_after(5, async_value.eventual_values()):
            ...
    """

    __slots__ = ['_ait', '_timeout']

    def __init__(self, timeout, ait):
        self._ait = ait
        self._timeout = timeout

    async def __anext__(self):
        with trio.fail_after(self._timeout):
            return await self._ait.__anext__()



from collections import defaultdict
from contextlib import _GeneratorContextManager
from functools import wraps
from inspect import iscoroutinefunction
from typing import Type, Dict, List

import trio


class _AsyncFriendlyGeneratorContextManager(_GeneratorContextManager):
    """
    Hack contextlib._GeneratorContextManager so that resulting context
    manager can properly decorate async functions.
    """

    def __call__(self, func):
        if iscoroutinefunction(func):
            @wraps(func)
            async def inner(*args, **kwargs):
                with self._recreate_cm():  # pylint: disable=not-context-manager
                    return await func(*args, **kwargs)
        else:
            @wraps(func)
            def inner(*args, **kwargs):
                with self._recreate_cm():  # pylint: disable=not-context-manager
                    return func(*args, **kwargs)
        return inner


def _async_friendly_contextmanager(func):
    """
    Equivalent to @contextmanager, except the resulting (non-async) context
    manager works correctly as a decorator on async functions.
    """
    @wraps(func)
    def helper(*args, **kwargs):
        return _AsyncFriendlyGeneratorContextManager(func, args, kwargs)
    return helper


def defer_to_cancelled(*args: Type[Exception]):
    """Context manager which defers MultiError exceptions to Cancelled.

    In the scope of this context manager, any raised trio.MultiError exception
    which is a combination of the given exception types and trio.Cancelled
    will have the exception types filtered, leaving only a Cancelled exception.

    The intended use is where routine exceptions (e.g. which are part of an API)
    might occur simultaneously with Cancelled (e.g. when using move_on_after()).
    Without properly catching and filtering the resulting MultiError, an
    unhandled exception will occur.  Often what is desired in this case is
    for the Cancelled exception alone to propagate to the cancel scope.

    Equivalent to ``multi_error_defer_to(trio.Cancelled, *args)``.

    :param args:  One or more exception types which will defer to
        trio.Cancelled.  By default, all exception types will be filtered.

    Example::

        # If MultiError([Cancelled, Obstacle]) occurs, propagate only Cancelled
        # to the parent cancel scope.
        with defer_to_cancelled(Obstacle):
            try:
                # async call which may raise exception as part of API
                await advance(speed)
            except Obstacle:
                # handle API exception (unless Cancelled raised simultaneously)
                ...
    """
    return multi_error_defer_to(trio.Cancelled, *args)


@_async_friendly_contextmanager
def multi_error_defer_to(*privileged_types: Type[BaseException],
                         propagate_multi_error=True,
                         strict=True):
    """
    Defer a trio.MultiError exception to a single, privileged exception

    In the scope of this context manager, a raised MultiError will be coalesced
    into a single exception with the highest privilege if the following
    criteria is met:

        1. every exception in the MultiError is an instance of one of the given
           privileged types

    additionally, by default with strict=True:

        2. there is a single candidate at the highest privilege after grouping
           the exceptions by repr().  For example, this test fails if both
           ValueError('foo') and ValueError('bar') are the most privileged.

    If the criteria are not met, by default the original MultiError is
    propagated.  Use propagate_multi_error=False to instead raise a
    RuntimeError in these cases.

    Examples::

        multi_error_defer_to(trio.Cancelled, MyException)
            MultiError([Cancelled(), MyException()]) -> Cancelled()
            MultiError([Cancelled(), MyException(),
                        MultiError([Cancelled(), Cancelled())]]) -> Cancelled()
            MultiError([Cancelled(), MyException(), ValueError()]) -> *no change*
            MultiError([MyException('foo'), MyException('foo')]) -> MyException('foo')
            MultiError([MyException('foo'), MyException('bar')]) -> *no change*

        multi_error_defer_to(MyImportantException, trio.Cancelled, MyBaseException)
            # where isinstance(MyDerivedException, MyBaseException)
            #   and isinstance(MyImportantException, MyBaseException)
            MultiError([Cancelled(), MyDerivedException()]) -> Cancelled()
            MultiError([MyImportantException(), Cancelled()]) -> MyImportantException()

    :param privileged_types: exception types from highest priority to lowest
    :param propagate_multi_error: if false, raise a RuntimeError where a
        MultiError would otherwise be leaked
    :param strict: propagate MultiError if there are multiple output exceptions
        to chose from (i.e. multiple exceptions objects with differing repr()
        are instances of the privileged type).  When combined with
        propagate_multi_error=False, this case will raise a RuntimeError.
    """
    try:
        yield
    except trio.MultiError as root_multi_error:
        # flatten the exceptions in the MultiError, grouping by repr()
        multi_errors = [root_multi_error]
        errors_by_repr = {}  # exception_repr -> exception_object
        while multi_errors:
            multi_error = multi_errors.pop()
            for e in multi_error.exceptions:
                if isinstance(e, trio.MultiError):
                    multi_errors.append(e)
                    continue
                if not isinstance(e, privileged_types):
                    # not in privileged list
                    if propagate_multi_error:
                        raise
                    raise RuntimeError('Unhandled trio.MultiError') from root_multi_error
                errors_by_repr[repr(e)] = e
        # group the resulting errors by index in the privileged type list
        # priority_index -> exception_object
        errors_by_priority: Dict[int, List[BaseException]] = defaultdict(list)
        for e in errors_by_repr.values():
            for priority, privileged_type in enumerate(privileged_types):
                if isinstance(e, privileged_type):
                    errors_by_priority[priority].append(e)
        # the error (or one of the errors) of the most privileged type wins
        priority_errors = errors_by_priority[min(errors_by_priority)]
        if strict and len(priority_errors) > 1:
            # multiple unique exception objects at the same priority
            if propagate_multi_error:
                raise
            raise RuntimeError('Unhandled trio.MultiError') from root_multi_error
        raise priority_errors[0]



from ._async_value import AsyncValue


class AsyncBool(AsyncValue):
    """Boolean wrapper offering the ability to wait for a value or transition.

    Synopsis::

        >>> a = AsyncBool()
        >>> ...
        >>> a.value = True   # access underlying value
        >>> ...
        >>> await a.wait_value(False)  # wait for a specific value
        >>> ...
        >>> await a.wait_transition()  # wait for a transition (default: any)

    When using `wait_value()` and `wait_transition()`, note that the value may
    have changed again before the caller receives control.

    Other than the constructor value defaulting to False, this class is the
    same as AsyncValue.
    """

    def __init__(self, value=False):
        super().__init__(value)



from ._async_value import AsyncValue


class RepeatedEvent:
    """A repeated event that supports multiple listeners.

    RepeatedEvent supports both "unqueued" and "eventual consistency" uses:

        * unqueued - drop events while processing the previous one
        * eventual consistency - some events may be missed while processing the
          previous one, but receiving the latest event is ensured
    """

    def __init__(self):
        self._event = AsyncValue(0)

    def set(self):
        """Trigger an event"""
        self._event.value += 1

    async def wait(self):
        """Wait for the next event"""
        token = self._event.value
        await self._event.wait_value(lambda val: val > token)

    async def unqueued_events(self):
        """Unqueued event iterator

        The listener will miss an event if it's blocked processing the previous
        one.  This is effectively the same as the following manual loop::

        >>> while True:
        >>>     await event.wait()
        >>>     # do work...

        Typical usage::

            >>> event = RepeatedEvent()

            A task listens for events:

            >>> async for _ in event.unqueued_events():
            >>>    # do blocking work
            >>>    await trio.sleep(1)

            Another task triggers events:

            >>> event.set()    # trigger event
            >>> trio.sleep(0)  # listener will enter loop body
            >>> event.set()    # listener misses this event since it's still in the loop body
            >>> trio.sleep(2)
            >>> event.set()    # listener will enter loop body again
        """
        async for _ in self._event.transitions():
            yield

    async def events(self, *, repeat_last=False):
        """Event iterator with eventual consistency

        Use this iterator to coordinate some work whenever a collection
        or other stateful object is mutated.  Although you may miss intermediate
        states, you're ensured to eventually receive an event to process the most
        recent state.  (https://en.wikipedia.org/wiki/Eventual_consistency)

        :param repeat_last: if true, repeat the last position in the event
          stream.  If no event has been set yet it still yields immediately,
          representing the "start" position.

        Typical usage::

            >>> my_list = []
            >>> repeated_event = RepeatedEvent()

            Whenever your collection is mutated, call the `set()` method.

            >>> my_list.append('hello')
            >>> repeated_event.set()

            The listener to continually process the latest state is:

            >>> async for _ in repeated_event.events():
            >>>     await persist_to_storage(my_list)

            If you'd like to persist the initial state of the list (before any
            set() is called), use the `repeat_last=True` option.
        """
        token = self._event.value
        if repeat_last:
            token -= 1
        async for _ in self._event.eventual_values(lambda val: val > token):
            yield



from contextlib import asynccontextmanager
from functools import partial
from typing import Awaitable, Callable, AsyncIterator

import trio

from ._awaitables import _wait_and_call


@asynccontextmanager
async def move_on_when(fn: Callable[..., Awaitable],
                       *args, **kwargs) -> AsyncIterator[trio.CancelScope]:
    """Async context manager that exits if async fn(*args, **kwargs) returns.

    The context manager yields a trio.CancelScope.

    Synopsis::

        async with move_on_when(my_event.wait) as cancel_scope:
            cancel_scope.shield = True
            await ...

        # by this point either the body exited, my_event was triggered, or both
    """

    async with trio.open_nursery() as nursery:
        with trio.CancelScope() as cancel_scope:
            nursery.start_soon(_wait_and_call,
                               partial(fn, *args, **kwargs),
                               cancel_scope.cancel)
            yield cancel_scope
        nursery.cancel_scope.cancel()


@asynccontextmanager
async def run_and_cancelling(fn: Callable[..., Awaitable],
                             *args, **kwargs) -> AsyncIterator[None]:
    """Async context manager that runs async fn(*args, **kwargs) and cancels it at block exit.

    Synopsis::

        async with run_and_cancelling(my_background_fn, my_arg=10):
            await ...
            # now the block exits, and my_background_fn is cancelled if still running
    """

    async with trio.open_nursery() as nursery:
        nursery.start_soon(partial(fn, *args, **kwargs))
        yield
        nursery.cancel_scope.cancel()



from collections import namedtuple
from contextlib import contextmanager, ExitStack
from functools import partial
from typing import ContextManager, Callable, Any, TypeVar, overload

from ._async_value import AsyncValue


def _IDENTITY(x):
    return x


T_OUT = TypeVar('T_OUT')


@overload
def compose_values(**value_map: AsyncValue) -> ContextManager[AsyncValue]: ...
@overload
def compose_values(*, _transform_: Callable[[Any], T_OUT],
                   **value_map: AsyncValue) -> ContextManager[AsyncValue[T_OUT]]: ...
def compose_values(*, _transform_=None, **value_map):
    """Context manager providing a composite of multiple AsyncValues

    The composite object itself is an AsyncValue, with the `value` of each
    underlying object accessible as attributes on the composite `value`.

    `compose_values()` expects named AsyncValue instances to be provided as
    keyword arguments.  The attributes of the composite value will correspond
    to the given names.

    It's mostly an implementation detail, but the composite value type is a
    namedtuple.  Users should not write to the composite `value` attribute
    since it is exclusively managed by the context.

    Synopsis:

        >>> async_x, async_y = AsyncValue(-1), AsyncValue(10)
        >>>
        >>> with compose_values(x=async_x, y=async_y) as async_xy:
        >>>     result = await async_xy.wait_value(lambda val: val.x < 0 < val.y))
        >>>
        >>> result
        CompositeValue(x=-1, y=10)

    The `_transform_` parameter specifies an optional function to transform the
    final value.  This is equivalent but more efficient than chaining a single
    open_transform() to the default compose_values() output.  For example:

        >>> with compose_values(x=async_x, y=async_y,
        >>>                     _transform_=lambda val: val.x * val.y) as x_mul_y:
        >>>     ...

    is equivalent to:

        >>> with compose_values(x=async_x, y=async_y) as async_xy, \\
        >>>         async_xy.open_transform(lambda val: val.x * val.y) as x_mul_y:
        >>>     ...

    Performance note:  predicates on the output AsyncValue will be evaluated
    on every assignment to the `value` properties of the input AsyncValues.
    So if two inputs are being composed, each updated 10 times per second,
    the output predicates will be evaluated 20 times per second.
    """
    # type hint workaround for https://youtrack.jetbrains.com/issue/PY-36444
    return _compose_values(_transform_, value_map)


@contextmanager
def _compose_values(_transform_, value_map):
    transform = _transform_ or _IDENTITY
    async_vals = value_map.values()
    if not (async_vals and all(isinstance(av, AsyncValue) for av in async_vals)):
        raise TypeError('expected instances of AsyncValue')
    value_type = namedtuple('CompositeValue', value_map.keys())
    composite_value = value_type._make(av.value for av in async_vals)
    composite = AsyncValue(transform(composite_value))

    # This dummy wait_value() predicate hooks into each value and updates
    # the composite as a side effect.
    def _update_composite(name, val):
        nonlocal composite_value
        composite_value = composite_value._replace(**{name: val})
        composite.value = transform(composite_value)
        return False

    with ExitStack() as stack:
        for name_, async_val in value_map.items():
            # NOTE: by using AsyncValue internals we avoid running wait_value()
            # as a child task for each input.
            stack.enter_context(
                async_val._level_results.open_ref(partial(_update_composite, name_)))

        yield composite



import pathlib

from setuptools import setup

pkg_name = 'trio_util'
base_dir = pathlib.Path(__file__).parent
with open(base_dir / 'src' / pkg_name / '_version.py') as f:
    version_globals = {}
    exec(f.read(), version_globals)
    version = version_globals['__version__']

setup(
    name=pkg_name,
    description='Utility library for the Trio async/await framework',
    long_description='''
An assortment of utilities for the Trio async/await framework, including:

  * wait_any, wait_all, move_on_when - avoid nursery blocks for simple cases
  * AsyncBool, AsyncValue - value wrappers with the ability to wait for
    a specific value or transition
  * periodic - periodic loop which accounts for its own execution
    time
  * @trio_async_generator - decorator which adapts a generator containing
     Trio constructs for safe use
  * azip, azip_longest - async zip with parallel iteration
  * RepeatedEvent - if you really, really want to reuse an event
  * and more (... obscure stuff you probably don't need)!
''',
    long_description_content_type='text/markdown',
    version=version,
    author='GROOVE X, Inc.',
    author_email='gx-sw@groove-x.com',
    url='https://github.com/groove-x/trio-util',
    license='MIT',
    packages=[pkg_name],
    package_dir={'': 'src'},
    install_requires=[
        'async_generator',
        'trio >= 0.11.0'
    ],
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Framework :: Trio',
    ],
)



import trio


async def _wait_and_call(f1, f2):
    """await on f1() and call f2()"""
    await f1()
    f2()


async def wait_any(*args):
    """Wait until any of the given async functions are completed.

    Equivalent to creating a new nursery and calling `start_soon()` on
    each async function, where the first function to return will cause the
    nursery to be cancelled.

    If the function invocations require arguments, use partial()::

        await wait_any(partial(foo, 'hello'),
                       partial(bar, debug=True))
    """
    async with trio.open_nursery() as nursery:
        cancel = nursery.cancel_scope.cancel
        for f in args:
            nursery.start_soon(_wait_and_call, f, cancel)


async def wait_all(*args):
    """Wait until all of the given async functions are completed.

    Equivalent to creating a new nursery and calling `start_soon()` on
    each async function.

    NOTE: Be careful when using this with a function that returns when
    some non-permanent condition is satisfied (e.g. `AsyncBool.wait_value`).
    While waiting for the other async function to complete, the state which
    satisfied the condition may change.
    """
    async with trio.open_nursery() as nursery:
        for f in args:
            nursery.start_soon(f)



from collections import defaultdict
from contextlib import contextmanager


class _RefCountedDefaultDict(defaultdict):
    """defaultdict offering deterministic collection of unreferenced keys"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.refs_by_key = defaultdict(int)

    @contextmanager
    def open_ref(self, key):
        """Context manager that holds a reference to key, yielding the corresponding value

        At context exit, the given key will be deleted if there are no other
        open references.
        """
        self.refs_by_key[key] += 1
        try:
            yield self[key]
        finally:
            self.refs_by_key[key] -= 1
            if self.refs_by_key[key] == 0:
                del self[key]
                del self.refs_by_key[key]



import functools
import sys
from contextlib import asynccontextmanager

import trio
from async_generator import aclosing


def trio_async_generator(wrapped):
    """async generator pattern which supports Trio nurseries and cancel scopes

    Normally, it's not allowed to yield from a Trio nursery or cancel scope when
    implementing async generators.  This decorator makes it possible to do so,
    adapting a generator for safe use.

    Though the wrapped function is written as a normal async generator, usage
    of the wrapper is different:  the wrapper is an async context manager
    providing the async generator to be iterated.

    Synopsis::

        >>> @trio_async_generator
        >>> async def my_generator():
        >>>    # yield values, possibly from a nursery or cancel scope
        >>>    # ...
        >>>
        >>>
        >>> async with my_generator() as agen:
        >>>    async for value in agen:
        >>>        print(value)

    Implementation:  "The idea is that instead of pushing and popping the
    generator from the stack of the task that's consuming it, you instead run
    the generator code as a second task that feeds the consumer task values."
    See https://github.com/python-trio/trio/issues/638#issuecomment-431954073

    ISSUE: pylint is confused by this implementation, and every use will
      trigger not-async-context-manager
    """
    @asynccontextmanager
    @functools.wraps(wrapped)
    async def wrapper(*args, **kwargs):
        send_channel, receive_channel = trio.open_memory_channel(0)
        async with trio.open_nursery() as nursery:
            async def adapter():
                async with send_channel, aclosing(wrapped(*args, **kwargs)) as agen:
                    while True:
                        try:
                            # Advance underlying async generator to next yield
                            value = await agen.__anext__()
                        except StopAsyncIteration:
                            break
                        while True:
                            try:
                                # Forward the yielded value into the send channel
                                try:
                                    await send_channel.send(value)
                                except trio.BrokenResourceError:
                                    return
                                break
                            except BaseException:  # pylint: disable=broad-except
                                # If send_channel.send() raised (e.g. Cancelled),
                                # throw the raised exception back into the generator,
                                # and get the next yielded value to forward.
                                try:
                                    value = await agen.athrow(*sys.exc_info())
                                except StopAsyncIteration:
                                    return

            nursery.start_soon(adapter, name=wrapped)
            async with receive_channel:
                yield receive_channel
    return wrapper
